@@ -6,10 +6,17 @@ import uuid
 import xml.etree.ElementTree as ET
 import xmltodict
 
-from six import text_type, binary_type
+from six import text_type
 
 from winrm.transport import Transport
-from winrm.exceptions import WinRMError, WinRMOperationTimeoutError
+from winrm.exceptions import WinRMError, WinRMTransportError, WinRMOperationTimeoutError
+
+xmlns = {
+    'soapenv': 'http://www.w3.org/2003/05/soap-envelope',
+    'soapaddr': 'http://schemas.xmlsoap.org/ws/2004/08/addressing',
+    'wsmanfault': "http://schemas.microsoft.com/wbem/wsman/1/wsmanfault"
+}
+
 
 class Protocol(object):
     """This is the main class that does the SOAP request/response logic. There
@@ -23,32 +30,51 @@ class Protocol(object):
 
     def __init__(
             self, endpoint, transport='plaintext', username=None,
-            password=None, realm=None, service=None, keytab=None,
-            ca_trust_path=None, cert_pem=None, cert_key_pem=None,
+            password=None, realm=None, service="HTTP", keytab=None,
+            ca_trust_path='legacy_requests', cert_pem=None, cert_key_pem=None,
             server_cert_validation='validate',
             kerberos_delegation=False,
             read_timeout_sec=DEFAULT_READ_TIMEOUT_SEC,
             operation_timeout_sec=DEFAULT_OPERATION_TIMEOUT_SEC,
             kerberos_hostname_override=None,
+            message_encryption='auto',
+            credssp_disable_tlsv1_2=False,
+            send_cbt=True,
+            proxy='legacy_requests',
         ):
         """
         @param string endpoint: the WinRM webservice endpoint
-        @param string transport: transport type, one of 'plaintext' (default), 'kerberos', 'ssl'  # NOQA
+        @param string transport: transport type, one of 'plaintext' (default), 'kerberos', 'ssl', 'ntlm', 'credssp'  # NOQA
         @param string username: username
         @param string password: password
         @param string realm: unused
         @param string service: the service name, default is HTTP
         @param string keytab: the path to a keytab file if you are using one
-        @param string ca_trust_path: Certification Authority trust path
+        @param string ca_trust_path: Certification Authority trust path. If server_cert_validation is set to 'validate':
+                                        'legacy_requests'(default) to use environment variables,
+                                        None to explicitly disallow any additional CA trust path
+                                        Any other value will be considered the CA trust path to use.
         @param string cert_pem: client authentication certificate file path in PEM format  # NOQA
         @param string cert_key_pem: client authentication certificate key file path in PEM format  # NOQA
-        @param string server_cert_validation: whether server certificate should be validated on Python versions that suppport it; one of 'validate' (default), 'ignore' #NOQA
+        @param string server_cert_validation: whether server certificate should be validated on Python versions that support it; one of 'validate' (default), 'ignore' #NOQA
         @param bool kerberos_delegation: if True, TGT is sent to target server to allow multiple hops  # NOQA
         @param int read_timeout_sec: maximum seconds to wait before an HTTP connect/read times out (default 30). This value should be slightly higher than operation_timeout_sec, as the server can block *at least* that long. # NOQA
         @param int operation_timeout_sec: maximum allowed time in seconds for any single wsman HTTP operation (default 20). Note that operation timeouts while receiving output (the only wsman operation that should take any significant time, and where these timeouts are expected) will be silently retried indefinitely. # NOQA
         @param string kerberos_hostname_override: the hostname to use for the kerberos exchange (defaults to the hostname in the endpoint URL)
+        @param bool message_encryption_enabled: Will encrypt the WinRM messages if set to True and the transport auth supports message encryption (Default True).
+        @param string proxy: Specify a proxy for the WinRM connection to use. 'legacy_requests'(default) to use environment variables, None to disable proxies completely or the proxy URL itself.
         """
 
+        try:
+            read_timeout_sec = int(read_timeout_sec)
+        except ValueError as ve:
+            raise ValueError("failed to parse read_timeout_sec as int: %s" % str(ve))
+
+        try:
+            operation_timeout_sec = int(operation_timeout_sec)
+        except ValueError as ve:
+            raise ValueError("failed to parse operation_timeout_sec as int: %s" % str(ve))
+
         if operation_timeout_sec >= read_timeout_sec or operation_timeout_sec < 1:
             raise WinRMError("read_timeout_sec must exceed operation_timeout_sec, and both must be non-zero")
 
@@ -65,7 +91,12 @@ class Protocol(object):
             server_cert_validation=server_cert_validation,
             kerberos_delegation=kerberos_delegation,
             kerberos_hostname_override=kerberos_hostname_override,
-            auth_method=transport)
+            auth_method=transport,
+            message_encryption=message_encryption,
+            credssp_disable_tlsv1_2=credssp_disable_tlsv1_2,
+            send_cbt=send_cbt,
+            proxy=proxy,
+        )
 
         self.username = username
         self.password = password
@@ -75,6 +106,7 @@ class Protocol(object):
         self.server_cert_validation = server_cert_validation
         self.kerberos_delegation = kerberos_delegation
         self.kerberos_hostname_override = kerberos_hostname_override
+        self.credssp_disable_tlsv1_2 = credssp_disable_tlsv1_2
 
     def open_shell(self, i_stream='stdin', o_stream='stdout stderr',
                    working_directory=None, env_vars=None, noprofile=False,
@@ -119,19 +151,22 @@ class Protocol(object):
             # TODO ensure that rsp:WorkingDirectory should be nested within rsp:Shell  # NOQA
             shell['rsp:WorkingDirectory'] = working_directory
             # TODO check Lifetime param: http://msdn.microsoft.com/en-us/library/cc251546(v=PROT.13).aspx  # NOQA
-            #if lifetime:
+            # if lifetime:
             #    shell['rsp:Lifetime'] = iso8601_duration.sec_to_dur(lifetime)
         # TODO make it so the input is given in milliseconds and converted to xs:duration  # NOQA
         if idle_timeout:
             shell['rsp:IdleTimeOut'] = idle_timeout
         if env_vars:
-            env = shell.setdefault('rsp:Environment', {})
+            # the rsp:Variable tag needs to be list of variables so that all
+            # environment variables in the env_vars dict are set on the shell
+            env = shell.setdefault('rsp:Environment', {}).setdefault('rsp:Variable', [])
             for key, value in env_vars.items():
-                env['rsp:Variable'] = {'@Name': key, '#text': value}
+                env.append({'@Name': key, '#text': value})
 
         res = self.send_message(xmltodict.unparse(req))
-        #res = xmltodict.parse(res)
-        #return res['s:Envelope']['s:Body']['x:ResourceCreated']['a:ReferenceParameters']['w:SelectorSet']['w:Selector']['#text']
+
+        # res = xmltodict.parse(res)
+        # return res['s:Envelope']['s:Body']['x:ResourceCreated']['a:ReferenceParameters']['w:SelectorSet']['w:Selector']['#text']
         root = ET.fromstring(res)
         return next(
             node for node in root.findall('.//*')
@@ -146,9 +181,9 @@ class Protocol(object):
         header = {
             '@xmlns:xsd': 'http://www.w3.org/2001/XMLSchema',
             '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
-            '@xmlns:env': 'http://www.w3.org/2003/05/soap-envelope',
+            '@xmlns:env': xmlns['soapenv'],
 
-            '@xmlns:a': 'http://schemas.xmlsoap.org/ws/2004/08/addressing',
+            '@xmlns:a': xmlns['soapaddr'],
             '@xmlns:b': 'http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd',
             '@xmlns:n': 'http://schemas.xmlsoap.org/ws/2004/09/enumeration',
             '@xmlns:x': 'http://schemas.xmlsoap.org/ws/2004/09/transfer',
@@ -204,32 +239,80 @@ class Protocol(object):
     def send_message(self, message):
         # TODO add message_id vs relates_to checking
         # TODO port error handling code
-        return self.transport.send_message(message)
+        try:
+            resp = self.transport.send_message(message)
+            return resp
+        except WinRMTransportError as ex:
+            try:
+                # if response is XML-parseable, it's probably a SOAP fault; extract the details
+                root = ET.fromstring(ex.response_text)
+            except Exception:
+                # assume some other transport error; raise the original exception
+                raise ex
 
-    def close_shell(self, shell_id):
+            fault = root.find('soapenv:Body/soapenv:Fault', xmlns)
+            if fault is not None:
+                fault_data = dict(
+                    transport_message=ex.message,
+                    http_status_code=ex.code
+                )
+                wsmanfault_code = fault.find('soapenv:Detail/wsmanfault:WSManFault[@Code]', xmlns)
+                if wsmanfault_code is not None:
+                    fault_data['wsmanfault_code'] = wsmanfault_code.get('Code')
+                    # convert receive timeout code to WinRMOperationTimeoutError
+                    if fault_data['wsmanfault_code'] == '2150858793':
+                        # TODO: this fault code is specific to the Receive operation; convert all op timeouts?
+                        raise WinRMOperationTimeoutError()
+
+                fault_code = fault.find('soapenv:Code/soapenv:Value', xmlns)
+                if fault_code is not None:
+                    fault_data['fault_code'] = fault_code.text
+
+                fault_subcode = fault.find('soapenv:Code/soapenv:Subcode/soapenv:Value', xmlns)
+                if fault_subcode is not None:
+                    fault_data['fault_subcode'] = fault_subcode.text
+
+                error_message = fault.find('soapenv:Reason/soapenv:Text', xmlns)
+                if error_message is not None:
+                    error_message = error_message.text
+                else:
+                    error_message = "(no error message in fault)"
+
+                raise WinRMError('{0} (extended fault data: {1})'.format(error_message, fault_data))
+
+    def close_shell(self, shell_id, close_session=True):
         """
         Close the shell
         @param string shell_id: The shell id on the remote machine.
          See #open_shell
+        @param bool close_session: If we want to close the requests's session.
+         Allows to completely close all TCP connections to the server.
         @returns This should have more error checking but it just returns true
          for now.
         @rtype bool
         """
-        message_id = uuid.uuid4()
-        req = {'env:Envelope': self._get_soap_header(
-            resource_uri='http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd',  # NOQA
-            action='http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete',
-            shell_id=shell_id,
-            message_id=message_id)}
+        try:
+            message_id = uuid.uuid4()
+            req = {'env:Envelope': self._get_soap_header(
+                resource_uri='http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd',  # NOQA
+                action='http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete',
+                shell_id=shell_id,
+                message_id=message_id)}
 
-        # SOAP message requires empty env:Body
-        req['env:Envelope'].setdefault('env:Body', {})
+            # SOAP message requires empty env:Body
+            req['env:Envelope'].setdefault('env:Body', {})
+
+            res = self.send_message(xmltodict.unparse(req))
+            root = ET.fromstring(res)
+            relates_to = next(
+                node for node in root.findall('.//*')
+                if node.tag.endswith('RelatesTo')).text
+        finally:
+            # Close the transport if we are done with the shell.
+            # This will ensure no lingering TCP connections are thrown back into a requests' connection pool.
+            if close_session:
+                self.transport.close_session()
 
-        res = self.send_message(xmltodict.unparse(req))
-        root = ET.fromstring(res)
-        relates_to = next(
-            node for node in root.findall('.//*')
-            if node.tag.endswith('RelatesTo')).text
         # TODO change assert into user-friendly exception
         assert uuid.UUID(relates_to.replace('uuid:', '')) == message_id
 
@@ -312,6 +395,39 @@ class Protocol(object):
         # TODO change assert into user-friendly exception
         assert uuid.UUID(relates_to.replace('uuid:', '')) == message_id
 
+    def send_command_input(self, shell_id, command_id, stdin_input, end=False):
+        """
+        Send input to the given shell and command.
+        @param string shell_id: The shell id on the remote machine.
+         See #open_shell
+        @param string command_id: The command id on the remote machine.
+         See #run_command
+        @param string stdin_input: The input unicode string or byte string to be sent.
+        @param bool end: Boolean value which will close the stdin stream. If end=True then the stdin pipe to the
+        remotely running process will be closed causing the next read by the remote process to stdin to return a
+        EndOfFile error; the behavior of each process when this error is encountered is defined by the process, but most
+        processes ( like CMD and powershell for instance) will just exit. Setting this value to 'True' means that no
+        more input will be able to be sent to the process and attempting to do so should result in an error.
+        @return: None
+        """
+        if isinstance(stdin_input, text_type):
+            stdin_input = stdin_input.encode("437")
+        req = {'env:Envelope': self._get_soap_header(
+            resource_uri='http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd',  # NOQA
+            action='http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Send',  # NOQA
+            shell_id=shell_id)}
+        stdin_envelope = req['env:Envelope'].setdefault('env:Body', {}).setdefault(
+            'rsp:Send', {}).setdefault('rsp:Stream', {})
+        stdin_envelope['@CommandId'] = command_id
+        stdin_envelope['@Name'] = 'stdin'
+        if end:
+            stdin_envelope['@End'] = "true"
+        else:
+            stdin_envelope['@End'] = "false"
+        stdin_envelope['@xmlns:rsp'] = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell'
+        stdin_envelope['#text'] = base64.b64encode(stdin_input)
+        self.send_message(xmltodict.unparse(req))
+
     def get_command_output(self, shell_id, command_id):
         """
         Get the Output of the given shell and command
@@ -320,9 +436,9 @@ class Protocol(object):
         @param string command_id: The command id on the remote machine.
          See #run_command
         #@return [Hash] Returns a Hash with a key :exitcode and :data.
-         Data is an Array of Hashes where the cooresponding key
+         Data is an Array of Hashes where the corresponding key
         #   is either :stdout or :stderr.  The reason it is in an Array so so
-         we can get the output in the order it ocurrs on
+         we can get the output in the order it occurs on
         #   the console.
         """
         stdout_buffer, stderr_buffer = [], []
@@ -333,7 +449,7 @@ class Protocol(object):
                     self._raw_get_command_output(shell_id, command_id)
                 stdout_buffer.append(stdout)
                 stderr_buffer.append(stderr)
-            except WinRMOperationTimeoutError as e:
+            except WinRMOperationTimeoutError:
                 # this is an expected error when waiting for a long-running process, just silently retry
                 pass
         return b''.join(stdout_buffer), b''.join(stderr_buffer), return_code
